/*****************************************************************************
 * $LastChangedDate: 2009-05-15 14:25:58 -0500 (Fri, 15 May 2009) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Module configuration.
 *//*
 * LEGAL:	COPYRIGHT (C) 2009 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define FX_CONF_CC 1
#include "base/module.hh"
#include "base/conf.hh"
#include "base/misc.hh"
using namespace base;
#include "fx/module.hh"
#include "fx/conf.hh"
#include "fx/defs.hh"
#include "fx/shadow_graph.hh"

namespace fx {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Conf  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Conf::Conf( void )
:	mValid(false),
	// cmd-line args
	mShadowMode(fx::defs::SHADOW_MODE_DEFAULT),
	// .conf file
    mShadowLightNum(1),
    mShadowLightPos(500.0f, 500.0f, 3000.0f, 0.0f),
    mShadowLightDir(0.70710678f, 0.0f, -0.70710678f),
    mShadowLightColor(1.0f, 1.0f, 1.0f, 1.0f)
	// Environment:
{
	// NOP
}

/*****************************************************************************
 * Initialize/read configuration.
 *****************************************************************************/
void
Conf::Init( int argc, char** argv, const string& confTxtFilename )
{

BEGIN_ONCE {

	mValid = true;  // assign this first as *_CONF may expand to GetConf() which asserts it

	ParseCmdlineArgs( argc, argv );   // cmdline args parsed prior to reading .conf file
	ReadConfTxt( confTxtFilename );   // read from textual .conf file
	ParseCmdlineArgs2( argc, argv );  // cmdline args that override .conf file

} END_ONCE

}

/*****************************************************************************
 * Parse cmdline args.
 *****************************************************************************/
void
Conf::ParseCmdlineArgs( int argc, char** argv )
{
	// NOP
}

void
Conf::ParseCmdlineArgs2( int argc, char** argv )
{
	for ( int i = 1; i < argc; ++i )
	{
		if ( STREQ( argv[i], "-no-shadows" )
          or STREQ( argv[i], "-fast" ) )
		{
			mShadowMode = fx::defs::SHADOW_MODE_DISABLED;
			CLOG << "Disabled shadows." << endl;
		}
		else if ( STREQ( argv[i], "-fast-shadows" ) )
		{
			mShadowMode = fx::defs::SHADOW_MODE_SAMPLE1;
			CLOG << "Enabled fast shadows." << endl;
		}
		else if ( STREQ( argv[i], "-soft-shadows" ) )
		{
			mShadowMode = fx::defs::SHADOW_MODE_SAMPLE4;
			CLOG << "Enabled soft shadows." << endl;
		}
	}
}

/*****************************************************************************
 * Read from .conf file.
 *****************************************************************************/
void
Conf::ReadConfTxt( const string& confTxtFilename )
{
	try
	{
		ModuleConfTxt confTxt;
		confTxt.Read( string(base::GetDataDir()) + string(confTxtFilename) );
	}
	catch (...)
	{
		CERROR << "ERROR: cannot read " << confTxtFilename << endl;
		EXIT( 1 );
	}
}

void
Conf::ModuleConfTxt::ProcessField( string& field, ifstream& conf )
{
	bool error = false;

	if ( field == "ShadowLightNum" )
    {
        conf >> FX_CONF.mShadowLightNum;
    }
	else if ( field == "ShadowLightPosition" )  // 4D
    {
        conf >> FX_CONF.mShadowLightPos[XX];
        conf >> FX_CONF.mShadowLightPos[YY];
        conf >> FX_CONF.mShadowLightPos[ZZ];
        conf >> FX_CONF.mShadowLightPos[WW];
    }
	else if ( field == "ShadowLightDirection" )  // 3D
    {
        conf >> FX_CONF.mShadowLightDir[XX];
        conf >> FX_CONF.mShadowLightDir[YY];
        conf >> FX_CONF.mShadowLightDir[ZZ];
    }
    else if ( field == "ShadowLightColor" )
	{
		fp r, g, b, a;
		conf >> r;
		conf >> g;
		conf >> b;
		conf >> a;
		FX_CONF.mShadowLightColor.Set( r, g, b, a );
	}
	
    // Errors?
	if ( error )
	{
		CERROR << "ERROR[conf]: parameter for " << field << " is invalid." << endl;
		EXIT( 1 );
	}
}

/*****************************************************************************
 * Print Conf configuration.
 *****************************************************************************/
ostream&
operator<<( ostream& strm, const Conf& conf )
{
	std::ios::fmtflags savedFlags = strm.flags();

	strm << "conf [shadow]:"                                                      << endl
	     << "shadow mode                 = " << conf.mShadowMode               ;//<< endl
	strm.flags(savedFlags);

	return strm;
}

} // namespace fx
