/*****************************************************************************
 * $LastChangedDate: 2009-05-15 16:33:17 -0500 (Fri, 15 May 2009) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Graph containing nodes that cast/receive shadows.
 *//*
 * LEGAL:	COPYRIGHT (C) 2009 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define FX_SHADOW_GRAPH_CC 1
#include "base/module.hh"
using namespace base;
#include "gfx/module.hh"
#include "gfx/osg.hh"
using namespace gfx;
#include "view/module.hh"
#include "view/node_masks.hh"
using namespace view;
#include "shader/module.hh"
#include "shader/shader_args.hh"
#include "shader/shader_factory.hh"
using namespace shader;
#include "fx/module.hh"
#include "fx/conf.hh"
#include "fx/shadow_graph.hh"
#include "fx/texture.hh"

#include <osg/Light>
#include <osg/LightSource>
#include <osgShadow/ShadowedScene>
#include <osgShadow/MinimalShadowMap>

namespace fx {

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  ShadowGraph  /////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor.
 *****************************************************************************/
ShadowGraph::ShadowGraph( RefPtr<GroupNode> parent )
:   mShadowedScene(new osgShadow::ShadowedScene),
    mShadowMap(new osgShadow::MinimalShadowMap)
#if LIMIT_SHADOW_CASTERS
   ,mShadowCasterNode(NULL)
#endif
{
    // osgShadow::ShadowedScene is a group node that is the root of the shadowed sub-graph.
    // osgShadow::ShadowMap is an algorithm ("shadow technique").
    parent->addChild( mShadowedScene.get() );

    // Initialize ShadowedScene (group node).
    mShadowedScene->setShadowTechnique( mShadowMap.get() );
    mShadowedScene->setReceivesShadowTraversalMask( view::defs::NODE_MASK_SHADOW_RECEIVER );
    mShadowedScene->setCastsShadowTraversalMask( view::defs::NODE_MASK_SHADOW_CASTER );

    // Initialize ShadowMap (algorithm/technique).
    // This program is using its own shaders.
    mShadowMap->setShadowTextureUnit( fx::defs::TEXTURE_UNIT_SHADOW );
    mShadowMap->setShadowVertexShader( NULL );
    mShadowMap->setShadowFragmentShader( NULL );
    mShadowMap->setMainVertexShader( NULL );
    mShadowMap->setMainFragmentShader( NULL );

    // Initialize an OpenGL hw light.
    // Though rendering is done by shaders,
    // the shaders use standard light uniforms as input.
    RefPtr<osg::LightSource> lightSource = new osg::LightSource;
    RefPtr<osg::Light> light = new osg::Light;
    light->setLightNum(  FX_CONF.mShadowLightNum );
    light->setPosition(  FX_CONF.mShadowLightPos );
    light->setDirection( FX_CONF.mShadowLightDir );
    light->setAmbient(   FX_CONF.mShadowLightColor );
    light->setDiffuse(   FX_CONF.mShadowLightColor );
    light->setSpecular(  FX_CONF.mShadowLightColor );
    lightSource->setReferenceFrame( osg::LightSource::RELATIVE_RF );
    lightSource->setLight( light.get() );
    mShadowedScene->addChild( lightSource.get() );
}

ShadowGraph::~ShadowGraph()
{
    // NOP
}

/*****************************************************************************
 * Attach node that will either cast or receive a shadow.
 *****************************************************************************/
void
ShadowGraph::AttachNode( RefPtr<Node> node, const NodeSort nodeSort )
{
    const bool shadowCaster = (nodeSort.GetAttribs().mVal) & (NodeSort::Attribs::SHADOW_CASTER);

    // Compute node mask that determines whether to cast or receive a shadow.
    uint mask = node->getNodeMask();
#if LIMIT_SHADOW_CASTERS
    // Limit to having one shadow caster else the shadow-map algorithm will malfunction or be slow.
    if ( shadowCaster )
    {
        // Unless passing back same caster.
        if ( (mShadowCasterNode != NULL) and (mShadowCasterNode.get() != node.get()) )
            DisableShadows( mShadowCasterNode );    // clear shadow node-mask in predecessor
        mShadowCasterNode = node;                   // replace with successor
    }
#endif
    mask = mask & ~(view::defs::NODE_MASK_SHADOW_CASTER | NodeSort::Attribs::SHADOW_RECEIVER);
    if ( shadowCaster )
    {
        mask |= view::defs::NODE_MASK_SHADOW_CASTER;
    	SetUniform<int>( UNIFORM_NAME_SHADOW_MODE, fx::defs::SHADOW_MODE_DISABLED, *node );
    }
    else
    {
        mask |= view::defs::NODE_MASK_SHADOW_RECEIVER;
    	SetUniform<int>( UNIFORM_NAME_SHADOW_MODE, FX_CONF.mShadowMode, *node );
    }
    node->setNodeMask( mask );

    // Attach node.
    mShadowedScene->addChild( node.get() );

    // Must apply shader directly to this node.
    // Ordinarily, nodes are placed under the branch node of the scene graph
    // that has the shader, but in this case, the node is attached to ShadowedScene instead.
    GET_SHADER_FACTORY().AttachShader( nodeSort.GetShaderName(), node );
}

/*****************************************************************************
 * Detach node.
 *****************************************************************************/
void
ShadowGraph::DetachNode( RefPtr<Node> node )
{
    mShadowedScene->removeChild( node.get() );

#if LIMIT_SHADOW_CASTERS
    // Let the sole shadow caster be replaced.
    if ( node.get() == mShadowCasterNode.get() )
        mShadowCasterNode = NULL;
#endif
}

/*****************************************************************************
 * Disable shadowing on a node (presumably a part of a model that is shadowed).
 *****************************************************************************/
void
ShadowGraph::DisableShadows( RefPtr<Node> node )
{
    // Change node mask.
    uint mask = node->getNodeMask();
    mask = mask & ~(view::defs::NODE_MASK_SHADOW_CASTER | NodeSort::Attribs::SHADOW_RECEIVER);
    node->setNodeMask( mask );

    // Change uniform.
   	SetUniform<int>( UNIFORM_NAME_SHADOW_MODE, fx::defs::SHADOW_MODE_DISABLED, *node );
}

} // namespace fx
