/*****************************************************************************
 * $LastChangedDate: 2009-05-15 17:20:33 -0500 (Fri, 15 May 2009) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	SceneGraph class.
 *//*
 * LEGAL:	COPYRIGHT (C) 2009 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define GRAPH_SCENE_GRAPH_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "gfx/module.hh"
#include "gfx/rgba.hh"
#include "gfx/osg.hh"
using namespace gfx;
#include "object/module.hh"
#include "object/events.hh"
using namespace object;
#include "view/module.hh"
#include "view/node_masks.hh"
using namespace view;
#include "shader/module.hh"
#include "shader/shader_factory.hh"
using namespace shader;
#include "fx/module.hh"
#include "fx/shadow_graph.hh"
using namespace fx;
#include "graph/module.hh"
#include "graph/conf.hh"
#include "graph/scene_graph.hh"

#include <osg/BlendFunc>
#include <osgUtil/Optimizer>
#include <osgDB/WriteFile>

namespace graph {

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  SceneGraph  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

SceneGraph* SceneGraph::msInstance;
NodeSort	SceneGraph::msBranchNodeSort;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
SceneGraph::SceneGraph( void )
:	mRootNode(new osg::ClearNode),  // ClearNode is a Group
	mBranchNode(msBranchNodeSort.CreateGroupNode()),  // create group with default state
	mNodeSortMap(),
	mObjects(),
    mShadowGraph(new fx::ShadowGraph(mRootNode.get()))
{
	// Attach branch node.
	mRootNode->addChild( mBranchNode.get() );

    // Set default state on root node.
    SetDefaultState( mRootNode.get() );

	// Set background color.
	SetBackgroundColor( GRAPH_CONF.mBackgroundColor );

	// Set default shader.
	GET_SHADER_FACTORY().AttachShader( ShaderName(), mRootNode.get() );

	// Unlisten() isn't necessary for singletons.
    EVENT_DESTROY_OBJECT.Listen( new DestroyObjectListener( *this ) );
}

SceneGraph::~SceneGraph()
{
	// NOP
}

/*****************************************************************************
 * (private subroutine)
 *****************************************************************************/
void
SceneGraph::SetDefaultState( RefPtr<Node> node )
{
    RefPtr<osg::StateSet> stateSet = node->getOrCreateStateSet();
    const osg::StateAttribute::OverrideValue override = osg::StateAttribute::OVERRIDE;

    // Disable hw lighting.
    stateSet->setMode( GL_LIGHTING, osg::StateAttribute::OFF | override );

    // Set default blend function (don't enable blending).
    RefPtr<osg::BlendFunc> blendFunc = new osg::BlendFunc;
    blendFunc->setFunction( osg::BlendFunc::SRC_ALPHA, osg::BlendFunc::ONE_MINUS_SRC_ALPHA );
    stateSet->setAttributeAndModes( blendFunc.get(), osg::StateAttribute::ON );

    // Enable back-face culling.
    stateSet->setMode( GL_CULL_FACE, osg::StateAttribute::ON | override );
}

/*****************************************************************************
 * (private subroutine)
 *****************************************************************************/
void
SceneGraph::AttachNode( RefPtr<Node> node, RefPtr<GroupNode> parent )
{
    // Ensure node is unattached.
    if ( EX( node->getNumParents() == 0 ) )
	{
		// Attach node.
		parent->addChild( node.get() );
	}

ASSERT( node->getNumParents() == 1 );
}

/*****************************************************************************
 * Attach node to scene-graph under branch node with special graphical-state.
 *****************************************************************************/
void
SceneGraph::AttachNode( RefPtr<Node> node, const NodeSort nodeSort )
{
	// Optimize this node's subgraph.
	Optimize( node, nodeSort );

  	// Attach node to either regular or shadowed scene-graph.
    const NodeSort::Attribs::Int attribs = nodeSort.GetAttribs().mVal;
    if ( UX( attribs & (NodeSort::Attribs::SHADOW_CASTER | NodeSort::Attribs::SHADOW_RECEIVER) ) )
    {
        // Shadowed graph.
        mShadowGraph->AttachNode( node, nodeSort );
    }
    else  // regular graph
    {
        // Try to map NodeSort to a cached group node.
        NodeSortMap::iterator iter = mNodeSortMap.find( nodeSort );
        if ( iter != mNodeSortMap.end() )
        {
        	// Was cached.
        	RefPtr<GroupNode> groupNode = iter->second;
        	AttachNode( node, groupNode );
        }
        else
        {
        	// Create group node.
        	RefPtr<GroupNode> groupNode = nodeSort.CreateGroupNode();

        	// Cache group node.
        	mNodeSortMap.insert( make_pair<>( nodeSort, groupNode ) );

        	// Attach node to branch.
        	AttachNode( node, groupNode );

            // Attach branch to root.
            AttachNode( groupNode.get(), mRootNode.get() );
        }
    }
}

/*****************************************************************************
 * Create a branch of the scene-graph at group node having special graphical-state.
 * The regular AttachNode() places nodes under a group with default graphical state.
 *****************************************************************************/
void
SceneGraph::AttachBranchNode( RefPtr<GroupNode> groupNode )
{
	AttachNode( groupNode.get(), mRootNode.get() );
}

/*****************************************************************************
 * Attach the node of an Object.
 * Implies that SceneGraph holds a reference to Object.
 * Uses Object::GetNodeSort() as a hint for state-sorting.
 *****************************************************************************/
void
SceneGraph::AttachObject( shptr<Object> object )
{
    AttachObject( object, object->GetNodeSort() );
}

/*****************************************************************************
 * Similar to AttachObject() but overrides Object::GetNodeSort().
 *****************************************************************************/
void
SceneGraph::AttachObject( shptr<Object> object, const NodeSort nodeSort )
{
	// Attach Object's node.
	AttachNode( object->GetRootNode(), nodeSort );

	// AttachObject() implies that SceneGraph holds a reference to Object.
	// The C++ core needs a container of Objects (so SceneGraph is it).
	mObjects.insert( object );
}

/*****************************************************************************
 * Detach node.
 * Ok to call if node is already detached.
 *****************************************************************************/
void
SceneGraph::DetachNode( RefPtr<Node> node )
{
    // A node should be attached to 0 or 1 group nodes.
    const uint parentCnt = node->getNumParents();
    if ( parentCnt == 1 )
    {
    	// Has one parent which should be an osg::Group.
        RefPtr<osg::Group> parent = node->getParent( 0 );  // idx=0
    #if DEBUG
        ASSERT( parent->asGroup() != NULL );            // should be a Group or derivative
        ASSERT( parent->removeChild( node.get() ) );    // remove() should return true
    #else
        parent->removeChild( node.get() );
    #endif

        // Ok if not attached.
        mShadowGraph->DetachNode( node );
    }
    else // parentCnt != 1
    {
        // Multiple parents shouldn't occur and isn't handled.
        // Objects may have shared nodes for 3D models but
        // those should be underneath unique group nodes (osg::MatrixTransform).
        ASSERT( parentCnt == 0 );
    }

    ASSERT( node->getNumParents() == 0 );  // should have zero parents
}

/*****************************************************************************
 * Detach user-defined branch node.
 * Ok to call if node is already detached.
 *****************************************************************************/
void
SceneGraph::DetachBranchNode( RefPtr<GroupNode> branchNode )
{
	DetachNode( branchNode.get() );
}

/*****************************************************************************
 * Detach Object.
 * Ok to call if Object is already detached.
 *****************************************************************************/
void
SceneGraph::DetachObject( shptr<Object> object )
{
	// Erasing is sufficient because ~Object() will detach its root node.
	mObjects.erase( object );
}

/*****************************************************************************
 * Run OSG optimizer on a node and its children.
 *****************************************************************************/
void
SceneGraph::Optimize( RefPtr<Node> node, const NodeSort nodeSort )
{
	if ( GRAPH_CONF.mEnableOsgOptimizer )
	{
	    osgUtil::Optimizer optimizer;
	    optimizer.optimize( node.get(), nodeSort.GetOptimizations().mVal );
	}
}

/*****************************************************************************
 * Optimize entire scene-graph.
 *****************************************************************************/
void
SceneGraph::Optimize( void )
{
	// This is both redundant and wrong.
	// Applying default optimizations would undo and break special models
	// that require specific NodeSort::Optimizations values.
	// However, this method remains as part of the gfxsys-neutral interface.
#if 0
	Optimize( GetRootNode().get(), NodeSort() );
#endif
}

/*****************************************************************************
 * Set the background color.
 *****************************************************************************/
void
SceneGraph::SetBackgroundColor( const RGBA color )
{
	// Root node is a osg::ClearNode.
	mRootNode->setClearColor( RGBA2VEC4F(color) );
}

/*****************************************************************************
 * Dump scene graph to file.
 *****************************************************************************/
void
SceneGraph::SaveToFile( const string& filename )
{
	COUT << "Saving scene-graph to file " << filename << "...";
	if ( osgDB::writeNodeFile( *GetRootNode(), filename ) )
		COUT << "ok." << endl;
	else
		CERROR << "ERROR" << endl;
}

/*****************************************************************************
 * Listen for the DestroyObject event.
 *****************************************************************************/
void
SceneGraph::DestroyObjectListener::operator()( shptr<Object> object )
{
CHECK_TYPESIG(object,TYPESIG_OBJECT);

	mSceneGraph.DetachObject( object );
}

} // namespace graph
