/*****************************************************************************
 * $LastChangedDate: 2009-05-15 14:25:58 -0500 (Fri, 15 May 2009) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	SceneGraph class.
 *//*
 * LEGAL:	COPYRIGHT (C) 2009 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef GRAPH_SCENE_GRAPH_HH
#define GRAPH_SCENE_GRAPH_HH 1

#define GET_SCENE_GRAPH() (SceneGraph::GetInstance())

#include "gfx/rgba.hh"
using namespace gfx;
#include "object/module.hh"
namespace object { class Object; }
#include "object/events.hh"
using namespace object;
namespace fx { class ShadowGraph; }
#include "graph/node_sort.hh"

namespace graph {

////////////////////////////////////////////////////////////////////////////////
/// @brief SceneGraph class contains nodes.
/// @verbatim
///                     +------------+
///                     | root node  |
///                 +---+------------+--+
///                 |        |          |
///                 |        |          |
///          +--------+  +--------+  +--------+
///          | branch |  | branch |  | branch |
///          |  node  |  |  node  |  |  node  |
///          | default|  |  [1]   |  |  [2]   |
///          +--------+  +--------+  +--------+
/// @endverbatim
///
/// This arrangement circumvents OSG's rule of allowing child nodes
/// to override similar state of its parent which wouldn't be restored
/// after traversal crosses over to a sibling node.
///
/// Different ways to attach a node:
/// --------------------------------
/// AttachNode( RefPtr<Node> node );
/// Attach node under branch node having default state.
///
/// AttachNode( RefPtr<Node> node, const NodeSort nodeSort );
/// Attach node under a branch node having state defined by NodeSort.
/// NodeSort defines typical graphical-state such as transparency.
/// SceneGraph creates and caches branch nodes mapped by NodeSort.
///
/// AttachBranchNode( RefPtr<GroupNode> branchNode );
/// For flexibility, allow client to externally create its own branch node
/// having special state.  Client would then attach geometry under it.
///
/// State-sorting responsibility:
/// -----------------------------
/// Nodes with a default graphical state can be attached to the root node.
/// If other graphical state is needed, the users of SceneGraph
/// are responsible for placing nodes sharing a graphical state
/// under a branch node which the user creates.
///
/// Shadows:
/// --------
/// Though not obvious, shadows are supported.
/// Attach shadow casters using NodeSort(SHADOW_CASTER)
/// and shadow receivers using NodeSort(SHADOW_RECEIVER).
///
class SceneGraph
{
PREVENT_COPYING(SceneGraph)
// Graph (ctor, dtor):
private:					SceneGraph( void );
private:					~SceneGraph();
public:
    CLASS_METHOD SceneGraph& GetInstance( void )  // Singleton
	{
		if ( UX( msInstance == NULL ) )
			msInstance = new SceneGraph;
		return *msInstance;
	}

// Graph (public interface):
// No methods accept Objects as SceneGraph is specifically for Nodes.
public:  RefPtr<GroupNode>	GetRootNode( void ) { return mRootNode.get(); }
public:	 void               AttachNode( RefPtr<Node> node, const NodeSort nodeSort = NodeSort() );
public:  void				AttachBranchNode( RefPtr<GroupNode> branchNode );
public:	 void               AttachObject( shptr<Object> object );
public:	 void               AttachObject( shptr<Object> object, const NodeSort nodeSort );
public:	 void				DetachNode( RefPtr<Node> node );
public:	 void				DetachBranchNode( RefPtr<GroupNode> branchNode );
public:	 void				DetachObject( shptr<Object> );
public:  void				Optimize( RefPtr<Node> node, const NodeSort nodeSort );
public:  void				Optimize( void ); 
public:	 void				SetBackgroundColor( const RGBA color );
public:  void				SaveToFile( const string& filename );

// Private methods:
private: void               SetDefaultState( RefPtr<Node> node );
private: void				AttachNode( RefPtr<Node> node, RefPtr<GroupNode> parent );
private: CLASS_DESTROY_OBJECT_LISTENER( SceneGraph, mSceneGraph )

// Graph (data members):
private: typedef map<NodeSort,RefPtr<GroupNode> > NodeSortMap;
private: typedef set<shptr<Object> > Objects;
private: CLASS_VAR SceneGraph*      msInstance;
private: CLASS_VAR NodeSort         msBranchNodeSort;	///< default branch has default NodeSort
private: RefPtr<osg::ClearNode>     mRootNode;			///< the absolute root node
private: RefPtr<osg::Group>         mBranchNode;		///< group node with default state
private: NodeSortMap                mNodeSortMap;		///< sorts nodes with shared state
private: Objects                    mObjects;			///< AttachObject() implies hold ref to Object
private: RefPtr<fx::ShadowGraph>    mShadowGraph;       ///< subgraph containing shadow casters/receivers
};

} // namespace graph

#endif // GRAPH_SCENE_GRAPH_HH
