/*****************************************************************************
 * $LastChangedDate: 2009-05-11 18:50:07 -0500 (Mon, 11 May 2009) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Hardware light for 3D models incompatible with shaders.
 *//*
 * LEGAL:	COPYRIGHT (C) 2009 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#if COMPILE_HW_LIGHT

#define SHADER_HW_LIGHT_CC 1
#include "base/module.hh"
using namespace base;
#include "gfx/module.hh"
#include "gfx/rgba.hh"
using namespace base;
#include "shader/module.hh"
#include "shader/conf.hh"
#include "shader/hw_light.hh"

#include <osg/Light>
#include <osg/LightSource>
#include <osg/LightModel>
#include <osg/Fog>

namespace shader {

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////  HwLight  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

HwLight* HwLight::msInstance;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
HwLight::HwLight( void )
:	mHwLight(new osg::Light),
    mHwFog(new osg::Fog),
    mAmbient(mHwLight->getAmbient()),  // -- ORDER-DEPENDENT --
    mDiffuse(mHwLight->getDiffuse()),
    mSpecular(mHwLight->getSpecular()),
    mFogColor(mHwFog->getColor())
{
    // Light:
    mHwLight->setLightNum( HW_LIGHT_NUM );

    // Fog:
	SetFogColor( SHADER_CONF.mFogColor );
	SetFogMode( SHADER_CONF.mFogMode );
	SetFogDensity( SHADER_CONF.mFogDensity );
}

HwLight::~HwLight()
{
	// NOP
}

/*****************************************************************************
 * Interface.
 * Create a new group node with hw light attached to it.
 * Every osg::LightSource (group node) shares the same osg::Light and osg::Fog.
 *****************************************************************************/
RefPtr<osg::LightSource>
HwLight::MakeHwLightGroup( void )
{
	// Every osg::LightSource (group node) shares the same osg::Light (mHwLight) and osg::Fog.
	RefPtr<osg::LightSource> lightGroup = new osg::LightSource;
    RefPtr<osg::StateSet> stateSet = lightGroup->getOrCreateStateSet();
    lightGroup->setLight( mHwLight.get() );
  //lightGroup->setReferenceFrame( osg::LightSource::ABSOLUTE_RF );
	lightGroup->setReferenceFrame( osg::LightSource::RELATIVE_RF );

	// Override OpenGL's base ambient light.
    RefPtr<osg::LightModel> lightModel = new osg::LightModel;
    lightModel->setAmbientIntensity( osg::Vec4(0,0,0,0) );
	lightGroup->getOrCreateStateSet()->setAttribute( lightModel.get() );

    // Attach fog.
	stateSet->setAttributeAndModes( mHwFog.get() );
	stateSet->setMode( GL_FOG, osg::StateAttribute::ON );

	return lightGroup.get();
}

/*****************************************************************************
 * Set brightness.
 * Brightness is changed by changing the color of hw light and hw fog.
 *****************************************************************************/
void
HwLight::SetBrightness( const fp brightness )
{
#define BRIGHTNESS_MUL_RGB(V4) osg::Vec4( (V4)[RR]*brightness, \
                                          (V4)[GG]*brightness, \
                                          (V4)[BB]*brightness, \
                                          (V4)[AA] )
    // Directly (bypassing saving to vars).
    mHwLight->setAmbient(  BRIGHTNESS_MUL_RGB(mAmbient) );
    mHwLight->setDiffuse(  BRIGHTNESS_MUL_RGB(mDiffuse) );
    mHwLight->setSpecular( BRIGHTNESS_MUL_RGB(mSpecular) );
    mHwFog->setColor(      BRIGHTNESS_MUL_RGB(mFogColor) );
}

/*****************************************************************************
 * Set colors of the default shader.
 *****************************************************************************/
void
HwLight::SetAmbient( const RGBA ambient )
{
    mAmbient = RGBA2VEC4F( ambient );  // save
    mHwLight->setAmbient( mAmbient );
}

void
HwLight::SetDiffuse( const RGBA diffuse )
{
    mDiffuse = RGBA2VEC4F( diffuse );  // save
    mHwLight->setDiffuse( mDiffuse );
}

void
HwLight::SetSpecular( const RGBA specular )
{
    mSpecular = RGBA2VEC4F( specular );  // save
    mHwLight->setSpecular( mSpecular );
}

/*****************************************************************************
 * Set position of light.
 *****************************************************************************/
void
HwLight::SetPosition( const WorldVertex position, const bool directional )
{
	const fp w = directional ? 0.0f : 1.0f;  // subtle OpenGL arg
	mHwLight->setPosition( osg::Vec4( position[XX], position[YY], position[ZZ], w ) );
}

/*****************************************************************************
 * Set fog color.
 *****************************************************************************/
void
HwLight::SetFogColor( const RGBA fogColor )
{
    mFogColor = RGBA2VEC4F( fogColor );  // save
	mHwFog->setColor( RGBA2VEC4F( mFogColor ) );
}

/*****************************************************************************
 * Set fog mode.
 *****************************************************************************/
void
HwLight::SetFogMode( const osg::Fog::Mode mode )
{
	mHwFog->setMode( mode );
}

/*****************************************************************************
 * Set fog density.
 *****************************************************************************/
void
HwLight::SetFogDensity( const fpx density )
{
	mHwFog->setDensity( density );
}

} // namespace shader

#endif // COMPILE_HW_LIGHT
